import isystem.connect as ic
import argparse

from typing import List
from datetime import datetime
from math import floor

'''
Disclaimer: This script needs 256 bytes of RAM available. It will perform dummy
read and write operations starting from specified RAM location onward.
'''


class Logger:
    def __init__(self, verbose: bool) -> None:
        ''' 
        Class constructor.
            @param verbose: If set to true, all messages will be outputted,
            including debug ones. 
        '''
        self.verbose = verbose

    def _log(self, message: str) -> None:
        '''
        Outputs formatted message to stdout.
        '''
        timestamp = datetime.now()
        timestampString = self.format_time(timestamp)

        formattedMessage = f"[{timestampString}]: {message}"
        print(formattedMessage)

    def debug(self, message: str) -> None:
        '''
        Function logs debug message to logger.
        '''
        if self.verbose:
            self._log(message)

    def log(self, message: str) -> None:
        '''
        Function logs user message to logger
        '''
        self._log(message)

    @staticmethod
    def format_time(time: datetime) -> str:
        '''
        Function formats specified time in more user-friendly form.
            @param time: Timestamp to be formatted.
        '''

        return time.strftime("%d.%m.%Y %H:%M:%S")


class DebugProtocol:
    def __init__(self, winIDEA: 'winIDEA') -> None:
        self.winIDEA = winIDEA

    def getDebugFrequencyOptionStr(self) -> List[str]:
        '''
        Retrieves winIDEA option string for setting debug frequency.
        '''
        return NotImplementedError

    def getString(self) -> str:
        '''
        Retrieves name of current debug protocol
        '''
        raise NotImplementedError

    def setDebugFrequency(self, frequency: int) -> None:
        '''
        Sets emulator debug frequency.
            @param frequency: Frequency in kHz
        '''

        debugFrequencies = self.getDebugFrequencyOptionStr()

        for debugFrequency in debugFrequencies:
            self.winIDEA.setOptionString(debugFrequency, str(frequency))


class JTAGProtocol(DebugProtocol):
    def getDebugFrequencyOptionStr(self) -> List[str]:
        return ["/iOPEN/Emulation.JTAG.ScanClock"]

    def getString(self) -> str:
        return "JTAG"


class SWDProtocol(DebugProtocol):
    def getDebugFrequencyOptionStr(self) -> List[str]:
        return ["/iOPEN/Emulation.OCD_ARM.CortexSWDClock"]

    def getString(self) -> str:
        return "SWD"


class DAPProtocol(DebugProtocol):
    def getDebugFrequencyOptionStr(self) -> List[str]:
        return [
            "/iOPEN/Emulation.OCD_TRICORE.DAPClock",
            "/iOPEN/Emulation.OCD_TRICORE.DAPE_Clock"
        ]

    def getString(self) -> str:
        return "DAP"


class LPDProtocol(DebugProtocol):
    def getDebugFrequencyOptionStr(self) -> List[str]:
        return ["/iOPEN/Emulation.OCD_V850.LPDClock"]

    def getString(self) -> str:
        return "LPD"


class winIDEA:
    def __init__(self, workspacePath: str) -> None:
        '''
        Class winIDEA constructor.
            @param workspacePath: Path to workspace where tests are performed.
        '''
        self.workspacePath = workspacePath

        self.cmgr = ic.ConnectionMgr()
        self.dbgCtrl = None
        self.workspaceCtrl = None
        self.dataCtrl = None
        self.ideCtrl = None
        self.execCtrl = None

    def open(self) -> None:
        '''
        Opens winIDEA and workspace which was provided in initializer.
        '''
        self.cmgr.connectMRU('')

        self.dbgCtrl = ic.CDebugFacade(self.cmgr)
        self.workspaceCtrl = ic.CWorkspaceController(self.cmgr)
        self.dataCtrl = ic.CDataController(self.cmgr)
        self.ideCtrl = ic.CIDEController(self.cmgr)
        self.execCtrl = ic.CExecutionController(self.cmgr)

        self.workspaceCtrl.open(self.workspacePath)

    def save(self) -> None:
        '''
        Saves winIDEA workspace.
        '''
        self.workspaceCtrl.save()

    def close(self) -> None:
        '''
        Closes winIDEA instance.
        '''
        self.cmgr.disconnect(ic.IConnect.dfCloseServerUnconditional,
                             ic.IConnect.dfCloseAutoSaveNone)

    def initTarget(self) -> None:
        ''' 
        Function initializes target. It performs reset and stops the CPU.
        '''
        self.dbgCtrl.reset()
        self.dbgCtrl.waitUntilStopped(1000)

    def getMemoryArea(self) -> int:
        '''
        Function retrieves memory area where data is stored.
        '''
        memoryArea = self.dataCtrl.getSystemMemoryAreas().getMemAreaDataVirtual()
        return memoryArea

    def getBytesPerMAU(self) -> int:
        '''
        Function retrieves minimal address unit size.
        '''
        memoryArea = self.getMemoryArea()
        bytesPerMau = self.dataCtrl.getMemoryAreaBytesPerMAU(memoryArea)
        return bytesPerMau

    def readMemory(self, address: int, length: int) -> List[int]:
        '''
        Reads value from specific address in target memory.
            @param address: Address from where value will be read
            @param length: Number of MAUs to be read
        '''

        readVal = self.dataCtrl.readMemory(ic.IConnectDebug.fMonitor,
                                           self.getMemoryArea(),
                                           address,
                                           length,
                                           self.getBytesPerMAU())

        value = list(readVal[0:length])

        return value

    def writeMemory(self, address: int, values: List[int]) -> None:
        '''
        Writes value to specific address in target memory.
            @param address: Address where value will be written to
            @param values: List of values to be written
        '''

        writeBuffer = ic.VectorBYTE(values)
        dataLength = len(values)

        self.dataCtrl.writeMemory(ic.IConnectDebug.fMonitor,
                                  self.getMemoryArea(),
                                  address,
                                  dataLength,
                                  self.getBytesPerMAU(),
                                  writeBuffer)

    def disconnect(self) -> None:
        '''
        Disconnects winIDEA debug session
        '''
        self.ideCtrl.iOpenDisconnect()

    def reattach(self) -> None:
        '''
        Function reestablishes debug session.
        '''
        self.disconnect()
        self.execCtrl.reset()

    def setOptionString(self, string: str, value: str) -> None:
        '''
        Function provides option string to IDE controller.
        '''
        self.ideCtrl.setOption(string, value)

    def getOptionString(self, string: str) -> str:
        '''
        Function gets option string value from IDE controller.
        '''
        value = self.ideCtrl.getOptionStr(string).lower()

        return value

    def _getCpuFamily(self) -> str:
        '''
        Function retrieves CPU family
        '''
        optionString = f"/iOPEN/Emulation.CPU.Family"

        cpuFamily = self.getOptionString(optionString)

        return cpuFamily

    def detectDebugProtocol(self) -> DebugProtocol:
        '''
        Function finds selected debug protocol in opened workspace.
        '''

        optionString = None
        cpuFamily = self._getCpuFamily()

        if 'tricore' in cpuFamily or '36' in cpuFamily:
            optionString = f"/iOPEN/Emulation.OCD_TRICORE.DebugMode"

        if 'arm' in cpuFamily or '21' in cpuFamily:
            optionString = f"/iOPEN/Emulation.OCD_ARM.CortexDebugProtocol"

        if 'rh850' in cpuFamily or '23' in cpuFamily:
            optionString = f"/iOPEN/Emulation.OCD_V850.DebugPort"

        if 'powerpc' in cpuFamily or '11' in cpuFamily:
            return JTAGProtocol(self)

        if optionString == None:
            raise Exception("Could not detect CPU architecture.")

        debugProtocol = self.getOptionString(optionString)

        if 'dap' in debugProtocol:
            return DAPProtocol(self)

        if 'swd' in debugProtocol:
            return SWDProtocol(self)

        if 'jtag' in debugProtocol:
            return JTAGProtocol(self)

        if 'lpd' in debugProtocol:
            return LPDProtocol(self)

        errorMsg = f"Unknown debug protocol detected: {debugProtocol}"

        raise Exception(errorMsg)

    def getEmulator(self) -> str:
        return self.getOptionString("/iOPEN/Hardware.Emulator")


def hexStr(number: int) -> str:
    '''
    Utility function which returns formatted hex string from an integer.
        @param number: Integer which will be formatted into hex string.
    '''
    return hex(number).upper().replace('X', 'x')


class APP:
    def __init__(self, workspacePath: str, ramLocation: int, debugFreqStart: int,
                 debugFreqEnd: int, threshold: int, logger: Logger) -> None:
        '''
        Class constructor.
            @param workspacePath: Path to workspace where tests are performed.
            @param ramLocation: Location of RAM as integer where RAM is started 
            in memory map.
            @param debugFreqStart: Minimal debug frequency to be tested.
            @param debugFreqEnd: Maximum debug frequency to be tested.
            @param threshold: Frequency threshold where test will be finished.
            @param logger: Logger instance for output messages
        '''

        if debugFreqEnd <= debugFreqStart:
            raise Exception("Maximum debug frequency should be higher than minimum debug frequency set.")

        if threshold:
            thresholdBisect = threshold
        else:
            thresholdBisect = (debugFreqEnd - debugFreqStart) / 100
            if thresholdBisect < 100:
                thresholdBisect = 100

        self.logger = logger
        self.winIDEA = winIDEA(workspacePath)
        self.bisect = Bisect(debugFreqStart, debugFreqEnd, thresholdBisect)

        self.ramLocation = ramLocation
        self.debugProtocol: DebugProtocol = None

        self.lastWorkingFrequency = 0

    def _showConfiguration(self) -> None:
        '''
        Function displays relevant informations about configuration to user.
        '''
        self.logger.log(f"Debug protocol: {self.debugProtocol.getString()}")
        self.logger.log(f"RAM location: {hexStr(self.ramLocation)}")
        self.logger.log(f"Minimum debug frequency: {self.bisect.getStartValue()} kHz")
        self.logger.log(f"Maximum debug frequency: {self.bisect.getEndValue()} kHz")
        self.logger.log(f"Frequency threshold: {self.bisect.getThreshold()} kHz")

    def _performTestAtFrequency(self, frequency: int) -> bool:
        '''
        Performs stress test at a given frequency.
            @param frequency: Frequency where the tests will be performed.
        '''
        try:
            self.debugProtocol.setDebugFrequency(frequency)
        except Exception as error:
            self.logger.log(f"Debug frequency cannot be set: {error}")
            return False

        try:
            self._performStressTest()

            self.logger.debug(f"Test at {frequency}kHz frequency: OK")
        except Exception as error:
            self.logger.debug(f"Error at {frequency}kHz frequency: {error}")
            return False

        return True

    def _performStressTest(self) -> None:
        '''
        Performs various memory read/write patterns to stress debug line.
        '''
        TEST_RETRIES = 10

        self.winIDEA.reattach()

        for _ in range(TEST_RETRIES):
            self.testRamReadWrite()

    def _applyConstraints(self) -> None:
        '''
        Applies frequency constraints on specific emulators.
        '''
        IC5000_MAX_FREQUENCY = 40000

        emulator = self.winIDEA.getEmulator()

        if emulator == 'ic5000':
            if self.bisect.getStartValue() > IC5000_MAX_FREQUENCY:
                raise Exception("Minimal frequency is too high for given emulator")

            if self.bisect.getEndValue() > IC5000_MAX_FREQUENCY:
                self.bisect.setEndValue(IC5000_MAX_FREQUENCY)
                self.logger.log(f"INFO: Maximum frequency was limited to {IC5000_MAX_FREQUENCY}kHz due to emulator capabilites.")

    def run(self) -> None:
        '''
        Main method for application
        '''

        try:
            self.winIDEA.open()

            self.debugProtocol = self.winIDEA.detectDebugProtocol()

            self.debugProtocol.setDebugFrequency(self.bisect.getStartValue())

            self._applyConstraints()

            self._showConfiguration()

            self.winIDEA.initTarget()
        except Exception as error:
            self.logger.log(f"Unexpected error during target initialization at minimal tested frequency. "
                            f"Tests will not be performed: {error}")
            return

        while True:
            frequency = self.bisect.get()

            result = self._performTestAtFrequency(frequency)

            if result:
                self.lastWorkingFrequency = frequency

            done = self.bisect.update(result)

            if not done:
                continue

            if self.lastWorkingFrequency == 0:
                self.logger.log(f"Bisection finished. Could not find maximum "
                                f"debug frequency.")
                break

            frequency = roundFrequency(self.lastWorkingFrequency)

            self.debugProtocol.setDebugFrequency(frequency)

            self.logger.log(f"Bisection finished. Maximum debug frequency "
                            f"is {frequency}kHz.")

            doSave = self._promptToSaveWorkspace()

            if doSave:
                self.winIDEA.save()
                self.logger.log("Workspace saved.")
            else:
                self.logger.log("Workspace is not saved.")

            break

        self.winIDEA.disconnect()

    def _promptToSaveWorkspace(self) -> bool:
        '''
        Prompts user to save modified workspace with newly set debug frequency.
        '''
        response = input("Do you want to save settings to your workspace (y/n)?")

        if "y" in response.lower():
            return True

        return False

    def _performStressTestWithData(self, writeCount: int, data: List[int]):
        '''
        Implementation of method which performs various memory reads and writes.
        It also checks whether memory write was performed correctly. Throws if
        memory write/read fails.
        '''
        dataSize = len(data)

        for offset in range(writeCount):
            address = self.ramLocation + offset * dataSize

            originalValue = self.winIDEA.readMemory(address, dataSize)
            self.winIDEA.writeMemory(address, data)

            readValue = self.winIDEA.readMemory(address, dataSize)
            self.winIDEA.writeMemory(address, originalValue)

            for idx in range(len(readValue)):
                if readValue[idx] != data[idx]:
                    localAddress = address + idx

                    msg = f"Error at address {hexStr(localAddress)}: " \
                        f"{hexStr(readValue[idx])} != {hexStr(data[idx])}"
                    raise Exception(msg)

    def testRamReadWrite(self) -> bool:
        '''
        This function performs dummy write and read operations in memory. Each
        element in DUMMY_DATA_PATTERN list must be provided with correct MAU
        unit.

        DUMMY_DATA_PATTERN is constructed in a way which is more likely to
        generate parity or frame error at higher frequencies. 

        Aurix: Data 0x3FA6'6666 generates CRC of 0x3F (all ones) which causes 
        can cause invalid start bit detection if line is not driven.  
        '''
        NUM_OF_WRITES = 16
        NUM_OF_WRITES_BULK = 1

        DUMMY_DATA_PATTERN = [0xAA, 0x55, 0xAA, 0x55, 0x3F, 0xA6, 0x66, 0x66]

        self._performStressTestWithData(NUM_OF_WRITES_BULK, DUMMY_DATA_PATTERN * NUM_OF_WRITES)
        self._performStressTestWithData(NUM_OF_WRITES, DUMMY_DATA_PATTERN)


class Bisect:
    def __init__(self, start: int, end: int, threshold: int) -> None:
        '''
        Class constructor.
            @params start: Bisection starting point
            @params end: Bisection end point.
            @params threshold: Threshold value which stops bisecting.
        '''
        self.start = start
        self.end = end

        self.current = start
        self.threshold = threshold

    def update(self, result: bool) -> bool:
        '''
        Updates bisection current value depending on result.
            @param result: Result of last operation in boolean
        '''
        last_current = self.current

        if result:
            if last_current == self.start:
                self.current = self.end
            else:
                self.start = self.current

                self.current = (self.end + self.start) // 2.0
        else:
            self.end = self.current

            self.current = (self.end + self.start) // 2.0

        if abs(last_current - self.get()) < self.threshold:
            return True

        return False

    def get(self) -> int:
        '''
        Retrieves next recommended value
        '''
        return int(self.current)

    def getStartValue(self) -> int:
        '''
        Retrieves bisection start value
        '''
        return int(self.start)

    def getEndValue(self) -> int:
        '''
        Retrieves bisection end value
        '''
        return int(self.end)

    def setEndValue(self, value: int) -> None:
        '''
        Sets end value for bisection
        '''
        self.end = value

    def getThreshold(self) -> int:
        '''
        Retrieves bisection threshold value
        '''
        return int(self.threshold)


def parse_arguments() -> argparse.Namespace:
    '''
    Function parses command line arguments and provides help about them.
    '''
    parser = argparse.ArgumentParser(description="Debug frequency bisection tester.")

    helpStr = "Path to workspace file."
    parser.add_argument("--workspace", "-w", required=True, help=helpStr)

    helpStr = "RAM location in memory. Value in HEX."
    parser.add_argument("--ramLocation", "-l", required=True, help=helpStr)

    helpStr = "Minimal debug frequency to be tested."
    parser.add_argument("--minFreq", "-m", required=False, help=helpStr, default=100)

    helpStr = "Maximum debug frequency to be tested."
    parser.add_argument("--maxFreq", "-M", required=False, help=helpStr, default=20000)

    helpStr = "Debug frequency threshold."
    parser.add_argument("--threshold", "-t", required=False, help=helpStr, default=0)

    helpStr = "Verbose output."
    parser.add_argument("--verbose", "-v", required=False, help=helpStr, action="store_true")

    options = parser.parse_args()

    return options


def convertToNumber(number: str) -> int:
    '''
    Converts string to number. Automatically detects number base.
    '''
    if isinstance(number, int):
        return number

    try:
        result = int(number, 16)
    except ValueError:
        result = int(number)
    return result


def roundFrequency(frequency: int) -> int:
    '''
    Returns frequency rounded down to the nearest 100.
    '''
    return int(floor(frequency / 100.0)) * 100


if __name__ == '__main__':
    options = parse_arguments()

    logger = Logger(verbose=options.verbose)

    app = APP(workspacePath=options.workspace,
              ramLocation=convertToNumber(options.ramLocation),
              debugFreqStart=int(options.minFreq),
              debugFreqEnd=int(options.maxFreq),
              threshold=int(options.threshold),
              logger=logger)

    app.run()
